class MaskData:

    def __init__(self):
        pass

    @classmethod
    def mask_headers(cls, headers, masks):
        """
        filters any headers (a dictionary) with key that matches those in the
        masks (a list)
        This have a side effect, so please make a deepcopy before using.
        """
        if headers is None:
            return headers
        if masks is None:
            return headers

        # Convert headers key to lowercase
        headers = {k.lower(): v for k, v in headers.items()}

        # Convert masks fields to lower case
        masks = [x.lower() for x in masks]

        for member in masks:
            headers.pop(member, None)

        return headers

    def mask_body(self, body, masks):
        """
        recursively removes any element from body (dictionary or lists) that
        have key matches masks. Note, this function have a side effect.
        Please make a deepcopy before using.
        """
        if body is None:
            return body
        if masks is None:
            return body

        if type(body) == list:
            return [self.mask_body(element, masks) for element in body]

        if type(body) == dict:
            for mask in masks:
                body.pop(mask, None)
            for key in body:
                body[key] = self.mask_body(body[key], masks)
            return body

        return body


from datetime import datetime
from .schemas import json_serialize

class JobScheduler:
    def __init__(self):
        pass

    @classmethod
    def exit_handler(cls, scheduler, debug):
        try:
            # Remove job and shut down the scheduler
            scheduler.remove_job('ironleap_events_batch_job')
            scheduler.shutdown()
        except Exception as ex:
            if debug:
                print("Error while closing the queue or scheduler shut down")
                print(str(ex))

    def send_events(self, client, batch_events, debug):
        try:
            if debug:
                print("Sending events to Iron Leap")
            client.post_events_batch(batch_events)
            if debug:
                print("Events sent successfully")
        except Exception as ex:
            if debug:
                print("Error sending event to Iron Leap")
                print(str(ex))

    def batch_events(self, client, events_queue, debug, batch_size):
        batch_events = []
        try:
            while not events_queue.empty():
                batch_events.append(events_queue.get_nowait())
                if len(batch_events) == batch_size:
                    break

            if batch_events:
                batch_response = self.send_events(client, batch_events, debug)
                batch_events[:] = []
                # Set the last time event job ran after sending events
                return batch_response, datetime.utcnow()
            else:
                if debug:
                    print("No events to send")
                # Set the last time event job ran but no message to read from the queue
                return None, datetime.utcnow()
        except Exception as e:
            if debug:
                print("No message to read from the queue")
                print(str(e))
            # Set the last time event job ran when exception occurred while sending event
            return None, datetime.utcnow()



import gzip
from .masks import MaskData
import json
import base64
try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO


class ParseBody:

    def __init__(self):
        self.mask_helper = MaskData()

    @classmethod
    def start_with_json(cls, body):
        return body.startswith("{") or body.startswith("[")

    @classmethod
    def transform_headers(cls, headers):
        return {k.lower(): v for k, v in headers.items()}

    @classmethod
    def base64_body(self, body):
        return base64.standard_b64encode(body).decode(encoding="UTF-8"), "base64"

    def parse_bytes_body(self, body, headers, body_masks):
        try:
            if (headers is not None and "content-encoding" in headers and headers["content-encoding"] is not None
                    and "gzip" in (headers["content-encoding"]).lower()):
                parsed_body, transfer_encoding = self.base64_body(gzip.decompress(body))
            else:
                string_data = body.decode(encoding="UTF-8")
                if self.start_with_json(string_data):
                    parsed_body = json.loads(string_data)
                    parsed_body = self.mask_helper.mask_body(parsed_body, body_masks)
                    transfer_encoding = 'json'
                else:
                    parsed_body, transfer_encoding = self.base64_body(body)
        except:
            parsed_body, transfer_encoding = self.base64_body(body)
        return parsed_body, transfer_encoding

    def parse_string_body(self, body, headers, body_masks):
        try:
            if self.start_with_json(body):
                parsed_body = json.loads(body)
                parsed_body = self.mask_helper.mask_body(parsed_body, body_masks)
                transfer_encoding = 'json'
            elif (headers is not None and "content-encoding" in headers and headers["content-encoding"] is not None
                  and "gzip" in (headers["content-encoding"]).lower()):
                decompressed_body = gzip.GzipFile(fileobj=StringIO(body)).read()
                parsed_body, transfer_encoding = self.base64_body(decompressed_body)
            else:
                parsed_body, transfer_encoding = self.base64_body(body)
        except:
            parsed_body, transfer_encoding = self.base64_body(body)
        return parsed_body, transfer_encoding



from __future__ import print_function

import django
import logging
import queue
from django.conf import settings
from django.utils import timezone
from io import BytesIO
from datetime import datetime, timedelta
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.interval import IntervalTrigger
import atexit
from .apiclient import ApiClient
from .job_scheduler import JobScheduler
from .schemas import APIEvent, APIRequest, APIResponse
from .logger_helper import LoggerHelper


class ironleap_middleware:
    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.
        self.middleware_settings = settings.IRONLEAP_MIDDLEWARE
        self.DEBUG = self.middleware_settings.get('DEBUG', False)
        self.LOG_BODY = self.middleware_settings.get('LOG_BODY', True)
        self.client = ApiClient(self.middleware_settings.get('APP_KEY'), self.middleware_settings.get('IRONLEAP_URL'))
        self.logger_helper = LoggerHelper()
        self.job_scheduler = JobScheduler()
        self.last_updated_time = datetime.utcnow()
        self.last_event_job_run_time = datetime(1970, 1, 1, 0, 0) # Assuming job never ran, set it to epoch start time
        self.scheduler = None
        self.event_queue_size = self.middleware_settings.get('EVENT_QUEUE_SIZE', 1000)
        self.events_queue = queue.Queue(maxsize=self.event_queue_size)
        self.event_batch_size = self.middleware_settings.get('BATCH_SIZE', 25)
        self.batch_send_interval = self.middleware_settings.get('BATCH_SEND_INTERVAL', 2)
        self.is_event_job_scheduled = False

    # Function to schedule send event job in async
    def schedule_event_background_job(self):
        try:
            if not self.scheduler:
                self.scheduler = BackgroundScheduler(daemon=True)
            if not self.scheduler.get_jobs():
                self.scheduler.start()
                self.scheduler.add_job(
                    func=lambda: self.job_scheduler.batch_events(
                        self.client,
                        self.events_queue,
                        self.DEBUG,
                        self.event_batch_size
                    ),
                    trigger=IntervalTrigger(seconds=2),
                    id='ironleap_events_batch_job',
                    name='Schedule events batch job every 2 second',
                    replace_existing=True,
                )
            
                # Avoid passing logging message to the ancestor loggers
                logging.getLogger('apscheduler.executors.default').setLevel(logging.WARNING)
                logging.getLogger('apscheduler.executors.default').propagate = False

                # Exit handler when exiting the app
                atexit.register(lambda: self.job_scheduler.exit_handler(self.scheduler, self.DEBUG))
        except Exception as ex:
            if self.DEBUG:
                print("Error when scheduling the job")
                print(str(ex))

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        # Setup response so it can be read from
        req_time = timezone.now()
        try:
            request._il_body = request.body
            request._stream = BytesIO(request.body)
            request._read_started = False
        except:
            request._il_body = None

        if self.DEBUG:
            print("executing get_response")
        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        # Safely parse request / response to create event to send to iron leap
        try:
            skip_event_response = self.logger_helper.skip_event(request, response, self.middleware_settings, self.DEBUG)
            if skip_event_response:
                return skip_event_response

            req_headers = self.logger_helper.parse_request_headers(request, self.middleware_settings, self.DEBUG)
            req_body, req_body_transfer_encoding = self.logger_helper.prepare_request_body(request, req_headers, self.LOG_BODY,
                                                                                        self.middleware_settings)
            uri = self.logger_helper.request_url(request)

            rsp_headers = self.logger_helper.parse_response_headers(response, self.middleware_settings)
            rsp_body, rsp_body_transfer_encoding = self.logger_helper.prepare_response_body(response, rsp_headers, self.LOG_BODY,
                                                                                            self.middleware_settings)
            rsp_time = timezone.now()

            company_id = self.logger_helper.get_company_id(self.middleware_settings, request, response, self.DEBUG)
            metadata = self.logger_helper.get_metadata(self.middleware_settings, request, response, self.DEBUG)

            event = APIEvent(
                request=APIRequest(
                    req_time.isoformat(),
                    uri,
                    request.method,
                    req_headers,
                    req_body,
                    req_body_transfer_encoding,
                ),
                response=APIResponse(
                    rsp_time.isoformat(),
                    response.status_code,
                    rsp_headers,
                    rsp_body,
                    rsp_body_transfer_encoding,
                ),
                company_id = company_id,
                metadata = metadata,
            )
            event = self.logger_helper.mask_event(event, self.middleware_settings, self.DEBUG)
        except Exception as ex:
            if self.DEBUG:
                print("Error while constructing event to send")
                print(str(ex))
            return response

        # Try adding event to queue
        try:
            if not self.is_event_job_scheduled and datetime.utcnow() > self.last_event_job_run_time + timedelta(minutes=5):
                try:
                    self.schedule_event_background_job()
                    self.is_event_job_scheduled = True
                    self.last_event_job_run_time = datetime.utcnow()
                except Exception as ex:
                    self.is_event_job_scheduled = False
                    if self.DEBUG:
                        print('Error while starting the event scheduler job in background')
                        print(str(ex))
            if self.DEBUG:
                print("Add APIEvent to the queue")
            self.events_queue.put(event)
        except Exception as ex:
            if self.DEBUG:
                print("Error while adding event to the queue")
                print(str(ex))

        return response



from datetime import datetime
import jsonpickle

class BaseModel(object):
    def to_dictionary(self):
        dictionary = dict()

        # Loop through all properties in this model
        for name in self.__dict__:
            value = getattr(self, name)
            if isinstance(value, list):
                # Loop through each item
                dictionary[name] = list()
                for item in value:
                    dictionary[name].append(item.to_dictionary() if isinstance(item, BaseModel) else item)
            elif isinstance(value, dict):
                # Loop through each item
                dictionary[name] = dict()
                for key in value:
                    dictionary[name][key] = value[key].to_dictionary() if isinstance(value[key], BaseModel) else value[key]
            elif isinstance(value, datetime):
                dictionary[name] = value.isoformat().to_dictionary() if isinstance(value, BaseModel) else value.isoformat()
            else:
                dictionary[name] = value.to_dictionary() if isinstance(value, BaseModel) else value
        # Return the result
        return dictionary

class APIEvent(BaseModel):
    def __init__(
        self,
        request = None,
        response = None,
        company_id = None,
        metadata = None
    ):
        self.request = request
        self.response = response
        self.company_id = company_id
        self.metadata = metadata

class APIRequest(BaseModel):
    def __init__(
        self,
        time = None,
        uri = None,
        verb = None,
        headers = None,
        body = None,
        transfer_encoding = None
    ):
        self.time = time
        self.uri = uri
        self.verb = verb
        self.headers = headers
        self.body = body
        self.transfer_encoding = transfer_encoding

class APIResponse(BaseModel):
    def __init__(
        self,
        time = None,
        status = None,
        headers = None,
        body = None,
        transfer_encoding = None
    ):
        self.time = time
        self.status = status
        self.headers = headers
        self.body = body
        self.transfer_encoding = transfer_encoding


def json_serialize(obj):
    """JSON Serialization of a given object.
    Args:
        obj (object): The object to serialise.
    Returns:
        str: The JSON serialized string of the object.
    """
    if obj is None:
        return None

    # Resolve any Names if it's one of our objects that needs to have this called on
    if isinstance(obj, list):
        value = list()
        for item in obj:
            if isinstance(item, BaseModel):
                value.append(item.to_dictionary())
            else:
                value.append(item)
        obj = value
    else:
        if isinstance(obj, BaseModel):
            obj = obj.to_dictionary()

    return jsonpickle.encode(obj, False)



from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
long_description = ''
if path.exists('README.md'):
    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()

setup(
    name='ironleapdjango',
    author='Iron Leap Inc.',
    author_email='founders@ironleap.io',
    version='1.0.0',
    description='Iron Leap Django SDK',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://ironleap.io',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=[
        "requests==2.20.0",
        "jsonpickle==0.7.1",
        "python-dateutil==2.5.3",
        "apscheduler==3.6.1"
    ],
)



import requests
from .schemas import json_serialize


class ApiClient(object):
    def __init__(self, app_key, uri):
        self.uri = uri
        self.headers = {
            'content-type': 'application/json; charset=utf-8',
            'X-Ironleap-Application-ID': app_key,
        }

    def post_events_batch(self, batch):
        resp = requests.post(self.uri, headers=self.headers, data=json_serialize(batch))
        if (resp.status_code < 200) or (resp.status_code > 208):
            raise Exception("HTTP response suggests error")



from .masks import MaskData
from .parse_body import ParseBody
from .schemas import json_serialize
from django.http import HttpRequest, HttpResponse
import json
import base64
import re
import copy
import uuid



class LoggerHelper:
    def __init__(self):
        self.regex_http_ = re.compile(r'^HTTP_.+$')
        self.regex_content_type = re.compile(r'^CONTENT_TYPE$')
        self.regex_content_length = re.compile(r'^CONTENT_LENGTH$')
        self.mask_helper = MaskData()
        self.parse_body = ParseBody()

    @classmethod
    def flatten_to_string(cls, value):
        if type(value) == str:
            return value
        if value is None:
            return ''
        return json_serialize(value)

    @classmethod
    def mapper(cls, response, key):
        return copy.deepcopy(response[key])

    def parse_request_headers(self, request, middleware_settings, debug):
        req_headers = {}
        regex_http_start = re.compile('^HTTP_')
        try:
            for header in request.META:
                if self.regex_http_.match(header) or self.regex_content_type.match(
                        header) or self.regex_content_length.match(header):
                    normalized_header = regex_http_start.sub('', header)
                    normalized_header = normalized_header.replace('_', '-')
                    req_headers[normalized_header] = request.META[header]
            req_headers = self.mask_helper.mask_headers(req_headers, middleware_settings.get('REQUEST_HEADER_MASKS'))
        except Exception as inst:
            if debug:
                print("error encountered while copying request header")
                print(inst)
            req_headers = {}

        if debug:
            print("about to print what is in meta %d " % len(request.META))
            for x in request.META:
                print(x, ':', request.META[x])
            print("about to print headers %d " % len(req_headers))
            for x in req_headers:
                print(x, ':', req_headers[x])

        req_headers = {k: self.flatten_to_string(v) for k, v in req_headers.items()}
        return req_headers

    def prepare_request_body(self, request, req_headers, log_body, middleware_settings):
        req_body = None
        req_body_transfer_encoding = None
        if log_body and request._il_body:
            if isinstance(request._il_body, str):
                req_body, req_body_transfer_encoding = self.parse_body.parse_string_body(request._il_body,
                                                                                         self.parse_body.transform_headers(
                                                                                             req_headers),
                                                                                         middleware_settings.get(
                                                                                             'REQUEST_BODY_MASKS'))
            else:
                req_body, req_body_transfer_encoding = self.parse_body.parse_bytes_body(request._il_body,
                                                                                        self.parse_body.transform_headers(
                                                                                            req_headers),
                                                                                        middleware_settings.get(
                                                                                            'REQUEST_BODY_MASKS'))
        return req_body, req_body_transfer_encoding

    def parse_response_headers(self, response, middleware_settings):
        # a little hacky, using _headers, which is intended as a private variable.
        # From Django 3.2 onwards, headers is available however
        headers = response._headers if hasattr(response, '_headers') else response.headers
        rsp_headers = {k: self.mapper(response, k) for k, v in headers.items()}
        return self.mask_helper.mask_headers(rsp_headers, middleware_settings.get('RESPONSE_HEADER_MASKS'))

    def prepare_response_body(self, response, rsp_headers, log_body, middleware_settings):
        rsp_body = None
        rsp_body_transfer_encoding = None
        if log_body and isinstance(response, HttpResponse) and response.content:
            if isinstance(response.content, str):
                rsp_body, rsp_body_transfer_encoding = self.parse_body.parse_string_body(response.content,
                                                                                         self.parse_body.transform_headers(
                                                                                             rsp_headers),
                                                                                         middleware_settings.get(
                                                                                             'RESPONSE_BODY_MASKS'))
            else:
                rsp_body, rsp_body_transfer_encoding = self.parse_body.parse_bytes_body(response.content,
                                                                                        self.parse_body.transform_headers(
                                                                                            rsp_headers),
                                                                                        middleware_settings.get(
                                                                                            'RESPONSE_BODY_MASKS'))
        return rsp_body, rsp_body_transfer_encoding

    @classmethod
    def request_url(cls, request):
        return request.scheme + "://" + request.get_host() + request.get_full_path()

    @classmethod
    def get_company_id(cls, middleware_settings, request, response, debug):
        company_id = None
        try:
            identify_company = middleware_settings.get('IDENTIFY_COMPANY', None)
            if identify_company is not None:
                company_id = identify_company(request, response)
        except:
            if debug:
                print("can not execute identify_company function, please check iron leap settings.")
        return company_id

    @classmethod
    def get_metadata(cls, middleware_settings, request, response, debug):
        metadata = None
        try:
            get_metadata = middleware_settings.get('GET_METADATA', None)
            if get_metadata is not None:
                metadata = get_metadata(request, response)
        except:
            if debug:
                print("can not execute get_metadata function, please check iron leap settings.")
        return metadata

    @classmethod
    def skip_event(cls, request, response, middleware_settings, debug):
        try:
            skip_event = middleware_settings.get('SKIP', None)
            if skip_event is not None:
                if skip_event(request, response):
                    return response
        except:
            if debug:
                print("Having difficulty executing skip_event function. Please check iron leap settings.")
            return None

    @classmethod
    def mask_event(cls, event, middleware_settings, debug):
        try:
            mask_event_model = middleware_settings.get('MASK_EVENT_MODEL', None)
            if mask_event_model is not None:
                event = mask_event_model(event)
        except:
            if debug:
                print("Can not execute mask_event_model function. Please check iron leap settings.")
        return event
